# -*- coding: utf-8 -*-
import unittest

from decimal import Decimal
from . import horror_fobj
from nose.plugins.attrib import attr
from nose.tools import assert_equal
from nose.plugins.skip import SkipTest

try:
    # Python 2.6 doesn't provide these functions
    from nose.tools import assert_is_instance, assert_greater_equal
except ImportError:
    from .shim26 import assert_is_instance, assert_greater_equal

from messytables import (CSVTableSet, StringType, HTMLTableSet,
                         ZIPTableSet, XLSTableSet, XLSXTableSet, PDFTableSet,
                         ODSTableSet, headers_guess, headers_processor,
                         offset_processor, DateType, FloatType,
                         IntegerType, BoolType, rowset_as_jts,
                         types_processor, type_guess, ReadError,
                         null_processor)
import datetime
stringy = type(u'')


class ReadCsvTest(unittest.TestCase):
    def test_utf8bom_lost(self):
        fh = horror_fobj('utf8bom.csv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        row = list(row_set)[0]
        assert_equal(row[0].value, "kitten")

    def test_read_simple_csv(self):
        fh = horror_fobj('simple.csv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        assert_equal(7, len(list(row_set)))
        row = list(row_set.sample)[0]
        assert_equal(row[0].value, 'date')
        assert_equal(row[1].value, 'temperature')

        for row in list(row_set):
            assert_equal(3, len(row))
            assert_equal(row[0].type, StringType())

    def test_read_complex_csv(self):
        fh = horror_fobj('complex.csv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        assert_equal(4, len(list(row_set)))
        row = list(row_set.sample)[0]
        assert_equal(row[0].value, 'date')
        assert_equal(row[1].value, 'another date')
        assert_equal(row[2].value, 'temperature')
        assert_equal(row[3].value, 'place')

        for row in list(row_set):
            assert_equal(4, len(row))
            assert_equal(row[0].type, StringType())

    def test_overriding_sniffed(self):
        # semicolon separated values
        fh = horror_fobj('simple.csv')
        table_set = CSVTableSet(fh, delimiter=";")
        row_set = table_set.tables[0]
        assert_equal(7, len(list(row_set)))
        row = list(row_set.sample)[0]
        assert_equal(len(row), 1)

    def test_read_head_padding_csv(self):
        fh = horror_fobj('weird_head_padding.csv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        offset, headers = headers_guess(row_set.sample)
        assert 11 == len(headers), headers
        assert_equal('1985', headers[1].strip())
        row_set.register_processor(headers_processor(headers))
        row_set.register_processor(offset_processor(offset + 1))
        data = list(row_set.sample)
        for row in row_set:
            assert_equal(11, len(row))
        value = data[1][0].value.strip()
        assert value == u'Gefäßchirurgie', value

    def test_read_head_offset_csv(self):
        fh = horror_fobj('simple.csv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        offset, headers = headers_guess(row_set.sample)
        assert_equal(offset, 0)
        row_set.register_processor(offset_processor(offset + 1))
        data = list(row_set.sample)
        assert_equal(int(data[0][1].value), 1)
        data = list(row_set)
        assert_equal(int(data[0][1].value), 1)

    @attr("slow")
    def test_read_type_guess_simple(self):
        fh = horror_fobj('simple.csv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        types = type_guess(row_set.sample)
        expected_types = [DateType("%Y-%m-%d"), IntegerType(), StringType()]
        assert_equal(types, expected_types)

        row_set.register_processor(types_processor(types))
        data = list(row_set)
        header_types = [c.type for c in data[0]]
        assert_equal(header_types, [StringType()] * 3)
        row_types = [c.type for c in data[2]]
        assert_equal(expected_types, row_types)

    def test_apply_null_values(self):
        fh = horror_fobj('null.csv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        types = type_guess(row_set.sample, strict=True)
        expected_types = [IntegerType(), StringType(), BoolType(),
                          StringType()]
        assert_equal(types, expected_types)

        row_set.register_processor(types_processor(types))
        data = list(row_set)
        # treat null as non empty text and 0 as non empty integer
        assert [x.empty for x in data[0]] == [False, False, False, False]
        assert [x.empty for x in data[1]] == [False, False, False, False]
        assert [x.empty for x in data[2]] == [False, False, True, True]
        assert [x.empty for x in data[3]] == [False, False, False, False]
        assert [x.empty for x in data[4]] == [False, False, False, True]
        assert [x.empty for x in data[5]] == [False, False, False, True]

        # we expect None for Integers and "" for empty strings in CSV
        assert [x.value for x in data[2]] == [3, "null", None, ""], data[2]

    def test_null_process(self):
        fh = horror_fobj('null.csv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        row_set.register_processor(null_processor(['null']))
        data = list(row_set)

        nones = [[x.value is None for x in row] for row in data]
        assert_equal(nones[0], [False, True, False, False])
        assert_equal(nones[1], [False, False, False, True])
        assert_equal(nones[2], [False, True, False, False])

        types = type_guess(row_set.sample, strict=True)
        expected_types = [IntegerType(), BoolType(), BoolType(),
                          BoolType()]
        assert_equal(types, expected_types)

        row_set.register_processor(types_processor(types))

        # after applying the types, '' should become None for int columns
        data = list(row_set)
        nones = [[x.value is None for x in row] for row in data]
        assert_equal(nones[0], [False, True, False, False])
        assert_equal(nones[1], [False, False, False, True])
        assert_equal(nones[2], [False, True, True, True])

    def test_read_encoded_csv(self):
        fh = horror_fobj('utf-16le_encoded.csv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        assert_equal(328, len(list(row_set)))
        row = list(row_set.sample)[0]
        assert_equal(row[1].value, 'Organisation_name')

    def test_long_csv(self):
        fh = horror_fobj('long.csv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        data = list(row_set)
        assert_equal(4000, len(data))

    def test_small_csv(self):
        fh = horror_fobj('small.csv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        data = list(row_set)
        assert_equal(1, len(data))

    def test_skip_initials(self):
        def rows(skip_policy):
            fh = horror_fobj('skip_initials.csv')
            table_set = CSVTableSet(fh,
                                    skipinitialspace=skip_policy)
            row_set = table_set.tables[0]
            return row_set

        def second(row):
            return row[1].value

        assert "goodbye" in list(map(second, rows(True)))
        assert "    goodbye" in list(map(second, rows(False)))

    def test_guess_headers(self):
        fh = horror_fobj('weird_head_padding.csv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        offset, headers = headers_guess(row_set.sample)
        row_set.register_processor(headers_processor(headers))
        row_set.register_processor(offset_processor(offset + 1))
        data = list(row_set)
        assert 'Frauenheilkunde' in data[9][0].value, data[9][0].value

        fh = horror_fobj('weird_head_padding.csv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        row_set.register_processor(headers_processor(['foo', 'bar']))
        data = list(row_set)
        assert 'foo' in data[12][0].column, data[12][0]
        assert 'Chirurgie' in data[12][0].value, data[12][0].value

    def test_read_encoded_characters_csv(self):
        fh = horror_fobj('characters.csv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        offset, headers = headers_guess(row_set.sample)
        row_set.register_processor(headers_processor(headers))
        row_set.register_processor(offset_processor(offset + 1))
        data = list(row_set)
        assert_equal(382, len(data))
        assert_equal(data[0][2].value, u'雲嘉南濱海國家風景區管理處')
        assert_equal(data[-1][2].value, u'沈光文紀念廳')


class ReadZipTest(unittest.TestCase):
    def test_read_simple_zip(self):
        fh = horror_fobj('simple.zip')
        table_set = ZIPTableSet(fh)
        row_set = table_set.tables[0]
        assert_equal(7, len(list(row_set)))
        row = list(row_set.sample)[0]
        assert_equal(row[0].value, 'date')
        assert_equal(row[1].value, 'temperature')

        for row in list(row_set):
            assert_equal(3, len(row))
            assert_equal(row[0].type, StringType())


class ReadTsvTest(unittest.TestCase):
    def test_read_simple_tsv(self):
        fh = horror_fobj('example.tsv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        assert_equal(141, len(list(row_set)))
        row = list(row_set.sample)[0]
        assert_equal(row[0].value, 'hour')
        assert_equal(row[1].value, 'expr1_0_imp')
        for row in list(row_set):
            assert_equal(17, len(row))
            assert_equal(row[0].type, StringType())


class ReadSsvTest(unittest.TestCase):
    def test_read_simple_ssv(self):
        # semicolon separated values
        fh = horror_fobj('simple.ssv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        assert_equal(7, len(list(row_set)))
        row = list(row_set.sample)[0]
        assert_equal(row[0].value, 'date')
        assert_equal(row[1].value, 'temperature')

        for row in list(row_set):
            assert_equal(3, len(row))
            assert_equal(row[0].type, StringType())


class ReadPsvTest(unittest.TestCase):
    def test_read_simple_psv(self):
        # pipe/vertical bar ("|") separated values
        fh = horror_fobj('simple.psv')
        table_set = CSVTableSet(fh)
        row_set = table_set.tables[0]
        assert_equal(29, len(list(row_set)))
        row = list(row_set.sample)[0]
        assert_equal(row[0].value, 'Year')
        assert_equal(row[1].value, 'New dwellings')

        for row in list(row_set):
            assert_equal(6, len(row))
            assert_equal(row[0].type, StringType())


class ReadODSTest(unittest.TestCase):
    def test_read_simple_ods(self):
        fh = horror_fobj('simple.ods')
        table_set = ODSTableSet(fh)
        assert_equal(1, len(table_set.tables))
        row_set = table_set.tables[0]
        row = list(row_set.sample)[0]
        assert_equal(row[0].value, 'Name')
        assert_equal(row[1].value, 'Age')
        assert_equal(row[2].value, 'When')
        total = 4
        for row in row_set.sample:
            total = total - 1
            assert 3 == len(row), row
        assert_equal(total, 0)

    @attr("slow")
    def test_read_large_ods(self):
        fh = horror_fobj('large.ods')
        table_set = ODSTableSet(fh)
        assert_equal(6, len(table_set.tables))
        row_set = table_set.tables[0]
        row = next(row_set.raw())
        assert len(row) == 16384, len(row)
        for row in row_set.sample:
            assert len(row) == 16384, len(row)

    def test_ods_version_4412(self):
        fh = horror_fobj('loffice-4.4.1.2.ods')
        table_set = ODSTableSet(fh)
        assert_equal(1, len(table_set.tables))
        row_set = table_set.tables[0]
        rows = row_set_to_rows(row_set)
        assert_equal(rows[0][0], 'Name')
        assert_equal(rows[1][0], 'Bob')
        assert_equal(rows[2][0], 'Jane')
        assert_equal(rows[3][0], 'Ian')

    def test_ods_read_past_blank_lines(self):
        fh = horror_fobj('blank_line.ods')
        table_set = ODSTableSet(fh)
        assert_equal(1, len(table_set.tables))
        row_set = table_set.tables[0]
        rows = row_set_to_rows(row_set)
        assert_equal(rows[0][0], 'Name')
        assert_equal(rows[1][0], 'Bob')
        assert_equal(rows[2][0], 'Jane')
        assert_equal(rows[3][0], 'Ian')

    def test_ods_read_all_supported_formats(self):
        fh = horror_fobj('ods_formats.ods')
        table_set = ODSTableSet(fh)
        assert_equal(3, len(table_set.tables))
        row_set = table_set.tables[0]
        rows = row_set_to_rows(row_set)
        assert_equal(rows[0][0], "Date")
        assert_equal(rows[1][0], "2014-11-11")
        assert_equal(rows[2][0], "2001-01-01")
        assert_equal(rows[3][0], '')
        # time formats
        assert_equal(rows[0][1], "Time")
        assert_equal(rows[1][1], "PT11H12M12S")
        assert_equal(rows[2][1], "PT00H00M12S")
        assert_equal(rows[4][1], 'PT27H17M54S')
        assert_equal(rows[5][1], "Other")
        # boolean
        assert_equal(rows[0][2], "Boolean")
        assert_equal(rows[1][2], 'true')
        assert_equal(rows[2][2], 'false')
        # Float
        assert_equal(rows[0][3], "Float")
        assert_equal(rows[1][3], '11.11')
        # Currency
        assert_equal(rows[0][4], "Currency")
        assert_equal(rows[1][4], '1 GBP')
        assert_equal(rows[2][4], '-10000 GBP')
        # Percentage
        assert_equal(rows[0][5], "Percentage")
        assert_equal(rows[1][5], '2')
        # int
        assert_equal(rows[0][6], "Int")
        assert_equal(rows[1][6], '3')
        assert_equal(rows[4][6], '11')
        # Scientific value is used but its notation is not
        assert_equal(rows[1][7], '100000')
        # Fraction
        assert_equal(rows[1][8], '1.25')
        # Text
        assert_equal(rows[1][9], "abc")

    def test_ods_read_all_supported_formats_casted(self):
        fh = horror_fobj('ods_formats.ods')
        table_set = ODSTableSet(fh)
        assert_equal(3, len(table_set.tables))
        row_set = table_set.tables[0]
        rows = cast_row_set_to_rows(row_set)
        date_format = "%d/%m/%Y"
        assert_equal(rows[0][0], "Date")
        assert_equal(rows[1][0].strftime(date_format), "11/11/2014")
        assert_equal(rows[2][0].strftime(date_format), "01/01/2001")
        assert_equal(rows[3][0], '')
        # time formats
        time_format = "%S:%M:%H"
        assert_equal(rows[0][1], "Time")
        assert_equal(rows[1][1].strftime(time_format), "12:12:11")
        assert_equal(rows[2][1].strftime(time_format), "12:00:00")
        assert_equal(rows[3][1], 0)
        assert_equal(rows[4][1], datetime.timedelta(hours=27,
                                                    minutes=17,
                                                    seconds=54))
        assert_equal(rows[5][1], "Other")
        # boolean
        assert_equal(rows[0][2], "Boolean")
        assert_equal(rows[1][2], True)
        assert_equal(rows[2][2], False)
        # Float
        assert_equal(rows[0][3], "Float")
        assert_equal(rows[1][3], Decimal('11.11'))
        # Currency
        assert_equal(rows[0][4], "Currency")
        assert_equal(rows[1][4], Decimal('1'))
        assert_equal(rows[2][4], Decimal('-10000'))
        # Percentage
        assert_equal(rows[0][5], "Percentage")
        assert_equal(rows[1][5], Decimal('0.02'))
        # int
        assert_equal(rows[0][6], "Int")
        assert_equal(rows[1][6], 3)
        assert_equal(rows[4][6], 11)
        # Scientific value is used but its notation is not
        assert_equal(rows[1][7], 100000)
        # Fraction
        assert_equal(rows[1][8], Decimal('1.25'))
        # Text
        assert_equal(rows[1][9], "abc")

    def test_ods_read_multi_line_cell(self):
        fh = horror_fobj('multilineods.ods')
        table_set = ODSTableSet(fh)
        row_set = table_set.tables[0]
        rows = row_set_to_rows(row_set)
        assert_equal(rows[0][0], '1\n2\n3\n4')


def row_set_to_rows(row_set):
    rows = []
    for row in row_set:
        rows.append([cell.value for cell in row])
    return rows


def cast_row_set_to_rows(row_set):
    rows = []
    for row in row_set:
        rows.append([cell.type.cast(cell.value) for cell in row])
    return rows


class XlsxBackwardsCompatibilityTest(unittest.TestCase):
    def test_that_xlsx_is_handled_by_xls_table_set(self):
        """
        Should emit a DeprecationWarning.
        """
        fh = horror_fobj('simple.xlsx')
        assert_is_instance(XLSXTableSet(fh), XLSTableSet)


class ReadXlsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.large_xlsx_table_set = XLSTableSet(   # TODO
            horror_fobj('large.xlsx'))

    def test_read_simple_xls(self):
        fh = horror_fobj('simple.xls')
        table_set = XLSTableSet(fh)
        assert_equal(1, len(table_set.tables))
        row_set = table_set.tables[0]
        first_row = list(row_set.sample)[0]
        third_row = list(row_set.sample)[2]

        assert_is_instance(first_row[0].value, stringy)
        assert_is_instance(first_row[1].value, stringy)
        assert_is_instance(first_row[2].value, stringy)

        assert_is_instance(third_row[0].value, datetime.datetime)
        assert_is_instance(third_row[1].value, float)
        assert_is_instance(third_row[2].value, stringy)

        assert_equal(first_row[0].value, 'date')
        assert_equal(first_row[1].value, 'temperature')
        assert_equal(first_row[2].value, 'place')

        assert_equal(third_row[0].value, datetime.datetime(2011, 1, 2, 0, 0))
        assert_equal(third_row[1].value, -1)
        assert_equal(third_row[2].value, 'Galway')

        for row in list(row_set):
            assert 3 == len(row), row

    # Right now we can't read even passwordless encrypted files - in future
    # would be good to be able to.
    def test_attempt_read_encrypted_no_password_xls(self):
        fh = horror_fobj('encrypted_no_password.xls')
        errmsg = "Can't read Excel file: XLRDError('Workbook is encrypted',)"
        try:
            XLSTableSet(fh)
        except ReadError as e:
            assert e.args[0] == errmsg
        else:
            assert False, "Did not raise Read Error"

    def test_read_head_offset_excel(self):
        fh = horror_fobj('simple.xls')
        table_set = XLSTableSet(fh)
        row_set = table_set.tables[0]
        offset, headers = headers_guess(row_set.sample)
        assert_equal(offset, 0)
        row_set.register_processor(offset_processor(offset + 1))
        data = list(row_set.sample)
        assert_equal(int(data[0][1].value), 1)
        data = list(row_set)
        assert_equal(int(data[0][1].value), 1)

    def test_read_simple_xlsx(self):
        fh = horror_fobj('simple.xlsx')
        table_set = XLSTableSet(fh)
        assert_equal(1, len(table_set.tables))
        row_set = table_set.tables[0]
        first_row = list(row_set.sample)[0]
        third_row = list(row_set.sample)[2]

        assert_is_instance(first_row[0].value, stringy)
        assert_is_instance(first_row[1].value, stringy)
        assert_is_instance(first_row[2].value, stringy)

        assert_is_instance(third_row[0].value, datetime.datetime)
        assert_is_instance(third_row[1].value, float)
        assert_is_instance(third_row[2].value, stringy)

        assert_equal(first_row[0].value, 'date')
        assert_equal(first_row[1].value, 'temperature')
        assert_equal(first_row[2].value, 'place')

        assert_equal(third_row[0].value, datetime.datetime(2011, 1, 2, 0, 0))
        assert_equal(third_row[1].value, -1.0)
        assert_equal(third_row[2].value, 'Galway')

        for row in list(row_set):
            assert 3 == len(row), row

    def test_large_file_report_sheet_has_11_cols_52_rows(self):
        table = self.large_xlsx_table_set['Report']
        num_rows = len(list(table))
        num_cols = len(list(table)[0])

        assert_equal(52, num_rows)
        assert_equal(11, num_cols)
        num_cells = sum(len(row) for row in table)
        assert_equal(num_rows * num_cols, num_cells)

    def test_large_file_data_sheet_has_11_cols_8547_rows(self):
        table = self.large_xlsx_table_set['data']
        num_rows = len(list(table))
        num_cols = len(list(table)[0])

        assert_equal(8547, num_rows)
        assert_equal(11, num_cols)
        num_cells = sum(len(row) for row in table)
        assert_equal(num_rows * num_cols, num_cells)

    def test_large_file_criteria_sheet_has_5_cols_12_rows(self):
        table = self.large_xlsx_table_set['criteria']
        num_rows = len(list(table))
        num_cols = len(list(table)[0])

        assert_equal(5, num_cols)
        assert_equal(12, num_rows)
        num_cells = sum(len(row) for row in table)
        assert_equal(num_rows * num_cols, num_cells)

    def test_read_type_know_simple(self):
        fh = horror_fobj('simple.xls')
        table_set = XLSTableSet(fh)
        row_set = table_set.tables[0]
        row = list(row_set.sample)[1]
        types = [c.type for c in row]
        assert_equal(types, [DateType(None), FloatType(), StringType()])

    def test_bad_first_sheet(self):
        # First sheet appears to have no cells
        fh = horror_fobj('problematic_first_sheet.xls')
        table_set = XLSTableSet(fh)
        tables = table_set.tables
        assert_equal(0, len(list(tables[0].sample)))
        assert_equal(1000, len(list(tables[1].sample)))


class ReadHtmlTest(unittest.TestCase):
    def test_read_real_html(self):
        fh = horror_fobj('html.html')
        table_set = HTMLTableSet(fh)
        row_set = table_set.tables[0]
        assert_equal(200, len(list(row_set)))
        row = list(row_set.sample)[0]
        assert_equal(row[0].value.strip(), 'HDI Rank')
        assert_equal(row[1].value.strip(), 'Country')
        assert_equal(row[4].value.strip(), '2010')

    def test_invisible_text_html(self):
        fh = horror_fobj('invisible_text.html')
        table_set = HTMLTableSet(fh)
        row_set = table_set.tables[0]
        assert_equal(4, len(list(row_set)))
        row = list(row_set.sample)[1]
        assert_equal(row[5].value.strip(), '1 July 1879')

    def test_read_span_html(self):
        fh = horror_fobj('rowcolspan.html')
        table_set = HTMLTableSet(fh)
        row_set = table_set.tables[0]

        magic = {}
        for y, row in enumerate(row_set):
            for x, cell in enumerate(row):
                magic[(x, y)] = cell.value

        tests = {(0, 0): '05',
                 (0, 2): '25',
                 (0, 3): '',
                 (1, 3): '36',
                 (1, 6): '66',
                 (4, 7): '79',
                 (4, 8): '89'}

        for test in tests:
            assert_equal(magic[test], tests[test])

    def test_that_outer_table_contains_nothing(self):
        fh = horror_fobj('complex.html')
        tables = {}
        for table in HTMLTableSet(fh).tables:
            tables[table.name] = table

        # outer_table should contain no meaningful data
        outer_table = list(tables['Table 2 of 2'])
        assert_equal(len(outer_table), 1)
        assert_equal(len(outer_table[0]), 1)
        assert_equal(outer_table[0][0].value.
                     replace(" ", "").
                     replace("\n", ""),
                     "headfootbody")

    def test_that_inner_table_contains_data(self):
        fh = horror_fobj('complex.html')
        tables = {}
        for table in HTMLTableSet(fh).tables:
            tables[table.name] = table

        inner_table = tables['Table 1 of 2']
        cell_values = []
        for row in inner_table:
            for cell in row:
                cell_values.append(cell.value)
        assert_equal(['head', 'body', 'foot'], cell_values)

    def test_rowset_as_schema(self):
        from io import BytesIO as sio
        ts = CSVTableSet(sio(b'''name,dob\nmk,2012-01-02\n'''))
        rs = ts.tables[0]
        jts = rowset_as_jts(rs).as_dict()
        assert_equal(jts['fields'], [
            {'type': 'string', 'id': 'name', 'label': 'name'},
            {'type': 'date', 'id': 'dob', 'label': 'dob'}])

    def test_html_table_name(self):
        fh = horror_fobj('html.html')
        table_set = HTMLTableSet(fh)
        assert_equal('Table 1 of 3', table_set.tables[0].name)
        assert_equal('Table 2 of 3', table_set.tables[1].name)
        assert_equal('Table 3 of 3', table_set.tables[2].name)


class ReadPdfTest(unittest.TestCase):
    def setUp(self):
        with horror_fobj('simple.pdf') as fh:
            try:
                PDFTableSet(fh)
            except ImportError:
                # Optional library isn't installed. Skip the tests.
                raise SkipTest(
                    "pdftables is not installed, skipping PDF tests")

    def test_read_simple_pdf(self):
        with horror_fobj('simple.pdf') as fh:
            table_set = PDFTableSet(fh)

        assert_equal(1, len(list(table_set.tables)))

        (table,) = table_set.tables
        rows = list(table)

        assert_greater_equal(len(rows), 1)

    def test_pdf_names(self):
        with horror_fobj('simple.pdf') as fh:
            table_set = PDFTableSet(fh)
        assert_equal('Table 1 of 1 on page 1 of 1',
                     table_set.tables[0].name)



import io
import re
import zipfile

from lxml import etree

from messytables.core import RowSet, TableSet, Cell
from messytables.types import (StringType, DecimalType,
                               DateType, BoolType, CurrencyType,
                               TimeType, PercentageType)


ODS_NAMESPACES_TAG_MATCH = re.compile(
    b"(<office:document-content[^>]*>)", re.MULTILINE)
ODS_TABLE_MATCH = re.compile(
    b".*?(<table:table.*?<\/.*?:table>).*?", re.MULTILINE)
ODS_TABLE_NAME = re.compile(b'.*?table:name=\"(.*?)\".*?')
ODS_ROW_MATCH = re.compile(
    b".*?(<table:table-row.*?<\/.*?:table-row>).*?", re.MULTILINE)

NS_OPENDOCUMENT_PTTN = u"urn:oasis:names:tc:opendocument:xmlns:%s"
NS_CAL_PTTN = u"urn:org:documentfoundation:names:experimental:calc:xmlns:%s"
NS_OPENDOCUMENT_TABLE = NS_OPENDOCUMENT_PTTN % "table:1.0"
NS_OPENDOCUMENT_OFFICE = NS_OPENDOCUMENT_PTTN % "office:1.0"

TABLE_CELL = 'table-cell'
VALUE_TYPE = 'value-type'
COLUMN_REPEAT = 'number-columns-repeated'
EMPTY_CELL_VALUE = ''

ODS_VALUE_TOKEN = {
    "float": "value",
    "date": "date-value",
    "time": "time-value",
    "boolean": "boolean-value",
    "percentage": "value",
    "currency": "value"
}

ODS_TYPES = {
    'float': DecimalType(),
    'date': DateType('%Y-%m-%d'),
    'boolean': BoolType(),
    'percentage': PercentageType(),
    'time': TimeType()
}


class ODSTableSet(TableSet):
    """
    A wrapper around ODS files. Because they are zipped and the info we want
    is in the zipped file as content.xml we must ensure that we either have
    a seekable object (local file) or that we retrieve all of the content from
    the remote URL.
    """

    def __init__(self, fileobj, window=None, **kw):
        '''Initialize the object.

        :param fileobj: may be a file path or a file-like object. Note the
        file-like object *must* be in binary mode and must be seekable (it will
        get passed to zipfile).

        As a specific tip: urllib2.urlopen returns a file-like object that is
        not in file-like mode while urllib.urlopen *does*!

        To get a seekable file you *cannot* use
        messytables.core.seekable_stream as it does not support the full seek
        functionality.
        '''
        if hasattr(fileobj, 'read'):
            # wrap in a StringIO so we do not have hassle with seeks and
            # binary etc (see notes to __init__ above)
            # TODO: rather wasteful if in fact fileobj comes from disk
            fileobj = io.BytesIO(fileobj.read())

        self.window = window

        zf = zipfile.ZipFile(fileobj).open("content.xml")
        self.content = zf.read()
        zf.close()

    def make_tables(self):
        """
            Return the sheets in the workbook.

            A regex is used for this to avoid having to:

            1. load large the entire file into memory, or
            2. SAX parse the file more than once
        """
        namespace_tags = self._get_namespace_tags()
        sheets = [m.groups(0)[0]
                  for m in ODS_TABLE_MATCH.finditer(self.content)]
        return [ODSRowSet(sheet, self.window, namespace_tags)
                for sheet in sheets]

    def _get_namespace_tags(self):
        match = re.search(ODS_NAMESPACES_TAG_MATCH, self.content)
        assert match
        tag_open = match.groups()[0]
        tag_close = b'</office:document-content>'
        return tag_open, tag_close


class ODSRowSet(RowSet):
    """ ODS support for a single sheet in the ODS workbook. Unlike
    the CSV row set this is not a streaming operation. """

    def __init__(self, sheet, window=None, namespace_tags=None):
        self.sheet = sheet

        self.name = "Unknown"
        m = ODS_TABLE_NAME.match(self.sheet)
        if m:
            self.name = m.groups(0)[0]

        self.window = window or 1000

        # We must wrap the XML fragments in a valid header otherwise iterparse
        # will explode with certain (undefined) versions of libxml2. The
        # namespaces are in the ODS file, and change with the libreoffice
        # version saving it, so get them from the ODS file if possible. The
        # default namespaces are an option to preserve backwards compatibility
        # of ODSRowSet.
        if namespace_tags:
            self.namespace_tags = namespace_tags
        else:
            namespaces = {
                "dc": u"http://purl.org/dc/elements/1.1/",
                "draw": NS_OPENDOCUMENT_PTTN % u"drawing:1.0",
                "number": NS_OPENDOCUMENT_PTTN % u"datastyle:1.0",
                "office": NS_OPENDOCUMENT_PTTN % u"office:1.0",
                "svg": NS_OPENDOCUMENT_PTTN % u"svg-compatible:1.0",
                "table": NS_OPENDOCUMENT_PTTN % u"table:1.0",
                "text": NS_OPENDOCUMENT_PTTN % u"text:1.0",
                "calcext": NS_CAL_PTTN % u"calcext:1.0",
            }

            ods_header = u"<wrapper {0}>"\
                .format(" ".join('xmlns:{0}="{1}"'.format(k, v)
                        for k, v in namespaces.iteritems())).encode('utf-8')
            ods_footer = u"</wrapper>".encode('utf-8')
            self.namespace_tags = (ods_header, ods_footer)

        super(ODSRowSet, self).__init__(typed=True)

    def raw(self, sample=False):
        """ Iterate over all rows in this sheet. """
        rows = ODS_ROW_MATCH.findall(self.sheet)

        for row in rows:
            row_data = []

            block = self.namespace_tags[0] + row + self.namespace_tags[1]
            partial = io.BytesIO(block)
            empty_row = True

            for action, element in etree.iterparse(partial, ('end',)):
                if element.tag != _tag(NS_OPENDOCUMENT_TABLE, TABLE_CELL):
                    continue

                cell = _read_cell(element)
                if empty_row is True and cell.value != EMPTY_CELL_VALUE:
                    empty_row = False

                repeat = element.attrib.get(
                    _tag(NS_OPENDOCUMENT_TABLE, COLUMN_REPEAT))
                if repeat:
                    number_of_repeat = int(repeat)
                    row_data += [cell] * number_of_repeat
                else:
                    row_data.append(cell)

            if empty_row:
                # ignore blank lines
                continue

            del partial
            yield row_data
        del rows


def _read_cell(element):
    cell_type = element.attrib.get(_tag(NS_OPENDOCUMENT_OFFICE, VALUE_TYPE))
    value_token = ODS_VALUE_TOKEN.get(cell_type, 'value')
    if cell_type == 'string':
        cell = _read_text_cell(element)
    elif cell_type == 'currency':
        value = element.attrib.get(_tag(NS_OPENDOCUMENT_OFFICE, value_token))
        currency = element.attrib.get(_tag(NS_OPENDOCUMENT_OFFICE, 'currency'))
        cell = Cell(value + ' ' + currency, type=CurrencyType())
    elif cell_type is not None:
        value = element.attrib.get(_tag(NS_OPENDOCUMENT_OFFICE, value_token))
        cell = Cell(value, type=ODS_TYPES.get(cell_type, StringType()))
    else:
        cell = Cell(EMPTY_CELL_VALUE, type=StringType())

    return cell


def _read_text_cell(element):
    children = element.getchildren()
    text_content = []
    for child in children:
        if child.text:
            text_content.append(child.text)
        else:
            text_content.append(EMPTY_CELL_VALUE)
    if len(text_content) > 0:
        cell_value = '\n'.join(text_content)
    else:
        cell_value = EMPTY_CELL_VALUE
    return Cell(cell_value, type=StringType())


def _tag(namespace, tag):
    return '{%s}%s' % (namespace, tag)



import sys
from datetime import datetime, time
import xlrd
from xlrd.biffh import XLRDError

from messytables.core import RowSet, TableSet, Cell, CoreProperties
from messytables.types import (StringType, IntegerType,
                               DateType, FloatType)
from messytables.error import ReadError
from messytables.compat23 import PY2

class InvalidDateError(Exception):
    pass

XLS_TYPES = {
    1: StringType(),
    # NB: Excel does not distinguish floats from integers so we use floats
    # We could try actual type detection between floats and ints later
    # or use the excel format string info - see
    # https://groups.google.com/forum/?fromgroups=#!topic/
    #  python-excel/cAQ1ndsCVxk
    2: FloatType(),
    3: DateType(None),
    # this is actually boolean but we do not have a boolean type yet
    4: IntegerType()
}


class XLSTableSet(TableSet):
    """An excel workbook wrapper object.
    """

    def __init__(self, fileobj=None, filename=None, window=None,
                 encoding=None, with_formatting_info=True, **kw):
        '''Initialize the tableset.

        :param encoding: passed on to xlrd.open_workbook function
            as encoding_override
        :param with_formatting_info: passed to xlrd to get font details of cells
        '''
        def get_workbook():
            try:
                return xlrd.open_workbook(
                    filename=filename,
                    file_contents=read_obj,
                    encoding_override=encoding,
                    formatting_info=with_formatting_info)
            except XLRDError as e:
                _, value, traceback = sys.exc_info()
                if PY2:
                   raise ReadError("Can't read Excel file: %r" % value, traceback)
                else:
                   raise ReadError("Can't read Excel file: %r" % value).with_traceback(traceback)
        '''Initilize the tableset.

        :param encoding: passed on to xlrd.open_workbook function
            as encoding_override
        :param with_formatting_info: whether xlrd should provide details
            of the cells contents (e.g. colour, borders, etc.
            Not sure what the behaviour of properties is with this turned off.
            Turning this on apparently may have memory implications in xlrd.

        The convoluted "try it with with_formatting_info, then try it without" is
        necessary because xlrd doesn't currently support getting this information
        from XLSX files. Workarounds include converting the XLSX document in LibreOffice.
        '''
        self.window = window

        if not filename and not fileobj:
            raise Exception('You must provide one of filename or fileobj')

        if fileobj:
            read_obj = fileobj.read()
        else:
            read_obj = None

        try:
            self.workbook = get_workbook()
        except NotImplementedError as e:
            if not with_formatting_info:
                raise
            else:
                with_formatting_info=False
                self.workbook = get_workbook()


    def make_tables(self):
        """ Return the sheets in the workbook. """
        return [XLSRowSet(name, self.workbook.sheet_by_name(name), self.window)
                for name in self.workbook.sheet_names()]


class XLSRowSet(RowSet):
    """ Excel support for a single sheet in the excel workbook. Unlike
    the CSV row set this is not a streaming operation. """

    def __init__(self, name, sheet, window=None):
        self.name = name
        self.sheet = sheet
        self.window = window or 1000
        super(XLSRowSet, self).__init__(typed=True)

    def raw(self, sample=False):
        """ Iterate over all rows in this sheet. Types are automatically
        converted according to the excel data types specified, including
        conversion of excel dates, which are notoriously buggy. """
        num_rows = self.sheet.nrows
        for rownum in range(min(self.window, num_rows) if sample else num_rows):
            row = []
            for colnum, cell in enumerate(self.sheet.row(rownum)):
                try:
                    row.append(XLSCell.from_xlrdcell(cell, self.sheet, colnum, rownum))
                except InvalidDateError:
                    raise ValueError("Invalid date at '%s':%d,%d" % (
                        self.sheet.name, colnum+1, rownum+1))
            yield row

class XLSCell(Cell):
    @staticmethod
    def from_xlrdcell(xlrd_cell, sheet, col, row):
        value = xlrd_cell.value
        cell_type = XLS_TYPES.get(xlrd_cell.ctype, StringType())
        if cell_type == DateType(None):
            if value == 0:
                raise InvalidDateError
            year, month, day, hour, minute, second = \
                xlrd.xldate_as_tuple(value, sheet.book.datemode)
            if (year, month, day) == (0, 0, 0):
                value = time(hour, minute, second)
            else:
                value = datetime(year, month, day, hour, minute, second)
        messy_cell = XLSCell(value, type=cell_type)
        messy_cell.sheet = sheet
        messy_cell.xlrd_cell = xlrd_cell
        messy_cell.xlrd_pos = (row, col)  # necessary for properties, note not (x,y)
        return messy_cell

    @property
    def topleft(self):
        return self.properties.topleft

    @property
    def properties(self):
        return XLSProperties(self)

class XLSProperties(CoreProperties):
    KEYS = ['bold', 'size', 'italic', 'font_name', 'strikeout', 'underline',
            'font_colour', 'background_colour', 'any_border', 'all_border',
            'richtext', 'blank', 'a_date', 'formatting_string']
    def __init__(self, cell):
        self.cell = cell
        self.merged = {}

    @property
    def xf(self):
        return self.cell.sheet.book.xf_list[self.cell.xlrd_cell.xf_index]

    @property
    def font(self):
        return self.cell.sheet.book.font_list[self.xf.font_index]

    @property
    def formatting(self):
        return self.cell.sheet.book.format_map[self.xf.format_key]

    @property
    def rich(self):
        """returns a tuple of character position, font number which starts at that position
        https://secure.simplistix.co.uk/svn/xlrd/trunk/xlrd/doc/xlrd.html?p=4966#sheet.Sheet.rich_text_runlist_map-attribute"""
        return self.cell.sheet.rich_text_runlist_map.get(self.cell.xlrd_pos, None)

    def raw_span(self, always=False):
        """return the bounding box of the cells it's part of.
         https://secure.simplistix.co.uk/svn/xlrd/trunk/xlrd/doc/xlrd.html?p=4966#sheet.Sheet.merged_cells-attribute"""
        row, col = self.cell.xlrd_pos
        for box in self.cell.sheet.merged_cells:
            rlo, rhi, clo, chi = box
            # note the high indexes are NOT inclusive!
            rhi = rhi - 1
            chi = chi - 1
            if row >= rlo and row <= rhi and col >= clo and col <= chi:
                return rlo, rhi, clo, chi
        if always:
            return (row, row, col, col)
        else:
            return None

    @property
    def topleft(self):
        span = self.raw_span()
        if span is None:
            return True  # is a single cell
        else:
            rlo, _, clo, _ = span
            return (rlo, clo) == self.cell.xlrd_pos

    def get_formatting_string(self):
        return self.formatting.format_str

    def get_a_date(self):
        return self.formatting.type == 1

    def get_richtext(self):  # TODO - get_rich_fragments
        return bool(self.rich)

    def get_bold(self):
        return self.font.weight > 500

    def get_size(self):
        """in pixels"""
        return self.font.height / 20.0

    def get_italic(self):
        return bool(self.font.italic)

    def get_font_name(self):
        return self.font.name

    def get_strikeout(self):
        return bool(self.font.struck_out)

    def get_underline(self):
        return self.font.underline_type > 0

    def get_font_colour(self):
        # TODO
        return self.font.color_index ## more lookup required

    def get_blank(self):
        """Note that cells might not exist at all.
           Behaviour for spanned cells might be complicated: hence this function"""
        return self.cell.value == ''

    def get_background_colour(self):
        return self.xf.background.background_color_index ## more lookup required

    def get_any_border(self):
        b = self.xf.border
        return b.top_line_style > 0 or b.bottom_line_style > 0 or \
               b.left_line_style > 0 or b.right_line_style > 0

    def get_all_border(self):
        b = self.xf.border
        return b.top_line_style > 0 and b.bottom_line_style > 0 and \
               b.left_line_style > 0 and b.right_line_style > 0




# -*- coding: utf-8 -*-

import unittest
from . import horror_fobj
from messytables.any import any_tableset
from messytables.error import NoSuchPropertyError
from nose.tools import (
    assert_equal,
    assert_false,
    assert_raises,
    assert_true)
import lxml.html

try:
    # Python 2.6 doesn't provide assert_is_instance
    from nose.tools import assert_is_instance, assert_greater_equal
except ImportError:
    from .shim26 import assert_is_instance, assert_greater_equal


class TestCellProperties(unittest.TestCase):
    def test_core_properties(self):
        csv = any_tableset(horror_fobj('simple.csv'), extension="csv")
        for table in csv.tables:
            for row in table:
                for cell in row:
                    cell.properties  # vague existence
                    assert_false('anything' in cell.properties)


class TestCoreProperties(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.html = any_tableset(horror_fobj('rowcolspan.html'),
                                extension="html")
        first_row = list(list(cls.html.tables)[0])[0]
        cls.real_cell = first_row[1]

    def test_properties_implements_in(self):
        assert_true('html' in self.real_cell.properties)
        assert_false('invalid' in self.real_cell.properties)

    def test_properties_implements_keys(self):
        assert(list(self.real_cell.properties.keys()))

    def test_properties_implements_items(self):
        assert(list(self.real_cell.properties.items()))

    def test_properties_implements_get(self):
        assert_equal('default', self.real_cell.properties.get(
            'not_in_properties', 'default'))
        assert_equal(None, self.real_cell.properties.get('not_in_properties'))


class TestExcelSpanRich(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.xls = any_tableset(horror_fobj('span_rich.xls'), extension='xls')
        cls.table = list(list(cls.xls.tables)[0])

    def test_basic_rich(self):
        assert self.table[4][1].properties['richtext']
        assert not list(self.table)[4][2].properties['richtext']
        assert self.table[4][1].value == 'bold and italic'

    def test_topleft(self):
        assert self.table[0][0].topleft, self.table[0][0].value
        assert self.table[1][1].topleft, self.table[1][1].value
        assert not self.table[1][2].topleft, self.table[1][2].value
        assert not self.table[2][2].topleft, self.table[2][2].value


class TestExcelProperties(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.xls = any_tableset(horror_fobj('excel_properties.xls'), extension='xls')
        rows = list(cls.xls.tables)[0]
        cls.first_cells = [list(row)[0] for row in rows]
        cls.properties = [x.properties for x in cls.first_cells]

    def test_cell_has_bold(self):
        assert_true('bold' in self.properties[0])
        assert_true(self.properties[0]['bold'])
        assert_false(self.properties[1]['bold'])

    def test_cell_has_italic(self):
        assert_true(self.properties[1]['italic'])
        assert_false(self.properties[0]['italic'])

    def test_cell_has_underline(self):
        assert_true(self.properties[2]['underline'])
        assert_false(self.properties[1]['underline'])

    def test_cell_size(self):
        assert_true(self.properties[9]['size'] > 20)
        assert_true(self.properties[10]['size'] < 8)
        assert_true(self.properties[0]['size'] == 10)

    def test_cell_has_borders(self):
        assert_false(self.properties[0]['any_border'])
        assert_false(self.properties[0]['all_border'])
        assert_true(self.properties[7]['any_border'])
        assert_false(self.properties[7]['all_border'])
        assert_true(self.properties[8]['any_border'])
        assert_true(self.properties[8]['all_border'])

    def test_cell_has_fontname(self):
        assert_true(self.properties[0]['font_name'] == 'Arial')

    def test_cell_has_strikeout(self):
        assert_true(self.properties[11]['strikeout'])
        assert_false(self.properties[0]['strikeout'])

    def test_blank_cells(self):
        assert_true(self.properties[12]['blank'])
        assert_false(self.properties[13]['blank'])

    def test_date(self):
        assert_false(self.properties[13]['a_date'])
        assert_true(self.properties[14]['a_date'])
        assert_equal(self.properties[14]['formatting_string'], r'DD/MM/YY')
        assert_equal(self.properties[15]['formatting_string'], r'QQ\ YY')
        assert_equal(self.properties[16]['formatting_string'], r'YYYY')
        assert_equal(self.properties[17]['formatting_string'], r'YYYY\ MMM')
        assert_equal(self.properties[18]['formatting_string'], r'D\ MMM\ YYYY')




class TestHtmlProperties(unittest.TestCase):
    # <td colspan='2'> would create one 'real' and one 'fake' cell
    @classmethod
    def setUpClass(cls):
        cls.html = any_tableset(horror_fobj('rowcolspan.html'),
                                extension="html")
        cls.first_row = list(list(cls.html.tables)[0])[0]
        cls.real_cell = cls.first_row[1]
        cls.fake_cell = cls.first_row[2]

    def test_real_cells_have_properties(self):
        assert_greater_equal(
            set(self.real_cell.properties.keys()),
            set(['_lxml', 'html'])
            )

    def test_real_cells_have_lxml_property(self):
        lxml_element = self.real_cell.properties['_lxml']
        assert_is_instance(lxml_element, lxml.etree._Element)
        assert_equal(b'<td colspan="2">06</td>',
                     lxml.html.tostring(lxml_element))

    def test_real_cell_has_a_colspan(self):
        assert_equal(self.real_cell.properties['colspan'], 2)

    def test_fake_cells_have_no_lxml_property(self):
        assert_raises(KeyError, lambda: self.fake_cell.properties['_lxml'])
        assert_raises(NoSuchPropertyError, lambda: self.fake_cell.properties['_lxml'])

    def test_real_cells_have_html_property(self):
        html = self.real_cell.properties['html']
        assert_is_instance(html, bytes)
        assert_equal(b'<td colspan="2">06</td>', html)

    def test_fake_cells_have_no_html_property(self):
        assert_raises(KeyError, lambda: self.fake_cell.properties['html'])

class TestBrokenColspans(unittest.TestCase):
    def setUp(self):
        self.html = any_tableset(horror_fobj("badcolspan.html"),
                                 extension="html")

    def test_first_row(self):
        first_row = list(list(self.html.tables)[0])[0]
        self.assertEqual([cell.properties['colspan'] for cell in first_row],
                         [1, 1, 1, 1])

    def test_second_row(self):
        second_row = list(list(self.html.tables)[0])[1]
        self.assertEqual([cell.properties['colspan'] for cell in second_row],
                         [1, 1, 1, 1])

    def test_third_row(self):
        third_row = list(list(self.html.tables)[0])[2]
        self.assertEqual([cell.properties['colspan'] for cell in third_row],
                         [1, 1, 1, 1])

    def test_fourth_row(self):
        fourth_row = list(list(self.html.tables)[0])[3]
        self.assertEqual([cell.properties['colspan'] for cell in fourth_row],
                         [1, 1, 1, 1])



from messytables.core import RowSet, TableSet, Cell, CoreProperties
import lxml.html
from collections import defaultdict
import html5lib
import xml.etree.ElementTree as etree


def fromstring(s):
    tb = html5lib.getTreeBuilder("lxml", implementation=etree)
    p = html5lib.HTMLParser(tb, namespaceHTMLElements=False)
    return p.parse(s)


class HTMLTableSet(TableSet):
    """
    A TableSet from a HTML document.
    """
    def __init__(self, fileobj=None, filename=None, window=None, **kw):

        if filename is not None:
            fh = open(filename, 'r')
        else:
            fh = fileobj
        if not fh:
            raise TypeError('You must provide one of filename or fileobj')

        self.htmltables = []
        root = fromstring(fh.read())

        # Grab tables that don't contain tables, remove from root, repeat.
        while True:
            dropped = False
            tables = root.xpath('//table[not(@messytable)]')
            if not tables:
                break
            for t in tables:
                if not t.xpath(".//table[not(@messytable)]"):
                    self.htmltables.append(t)
                    t.attrib['messytable'] = 'done'
                    dropped = True
            assert dropped, "Didn't find any tables not containing " + \
                "other tables. This is a bug."  # avoid infinite loops

    def make_tables(self):
        """
        Return a listing of tables (as HTMLRowSets) in the table set.
        """
        def rowset_name(rowset, table_index):
            return "Table {0} of {1}".format(table_index + 1,
                                             len(self.htmltables))

        return [HTMLRowSet(rowset_name(rowset, index), rowset)
                for index, rowset in enumerate(self.htmltables)]


def insert_blank_cells(row, blanks):
    """
    Given a list of values, insert blank cells at the indexes given by blanks
    The letters in these examples should really be cells.
    >>> insert_blank_cells(["a","e","f"],[1,2,3])
    ['a', <Cell(String:>, <Cell(String:>, <Cell(String:>, 'e', 'f']
    """
    # DISCUSS: option to repeat top-left of col/rowspan.
    # or to identify that areas are a single cell, originally.
    for i in blanks:
        row.insert(i, FakeHTMLCell())
    return row


class HTMLRowSet(RowSet):
    """
    A RowSet representing a HTML table.
    """
    def __init__(self, name, sheet, window=None):
        self.name = name
        self.sheet = sheet
        self.window = window or 1000
        super(HTMLRowSet, self).__init__()

    def in_table(self, els):
        """
        takes a list of xpath elements and returns only those
        whose parent table is this one
        """

        return [e for e in els
                if self.sheet in e.xpath("./ancestor::table[1]")]

    def raw(self, sample=False):
        def identify_anatomy(tag):
            # 0: thead, 1: tbody, 2: tfoot
            parts = ['.//ancestor::thead',
                     './/ancestor::tbody',
                     './/ancestor::tfoot']
            for i, part in enumerate(parts):
                if self.in_table(tag.xpath(part)):
                    return i
            return 2  # default to body

        blank_cells = defaultdict(list)  # ie row 2, cols 3,4,6: {2: [3,4,6]}
        allrows = sorted(self.in_table(self.sheet.xpath(".//tr")),
                         key=lambda tag: identify_anatomy(tag))
        # http://stackoverflow.com/questions/1915376/ - sorted() is stable.

        for r, row in enumerate(allrows):
            # TODO: handle header nicer - preserve the fact it's a header!
            html_elements = self.in_table(
                row.xpath('.//*[name()="td" or name()="th"]'))
            html_cells = [HTMLCell(source=cell) for cell in html_elements]

            """ at the end of this chunk, you have accurate blank_cells."""
            output_column = 0
            for html_cell in html_cells:
                assert type(r) == int
                while output_column in blank_cells[r]:
                    output_column += 1  # pass over col, doesn't exist in src

                rowspan = html_cell.properties['rowspan']
                colspan = html_cell.properties['colspan']

                x_range = range(output_column, output_column + colspan)
                y_range = range(r, r + rowspan)
                for x in x_range:
                    for y in y_range:
                        if (output_column, r) != (x, y):
                            # don't skip current cell
                            blank_cells[y].append(x)
                output_column += 1

            yield insert_blank_cells(html_cells, blank_cells[r])
            if sample and r == self.window:
                return
            del blank_cells[r]


class FakeHTMLCell(Cell):
    def __init__(self):
        super(FakeHTMLCell, self).__init__("")

    @property
    def topleft(self):
        """
        FakeHTMLCells are those which are not physically present in the HTML
        because of column or row spannning.

        See also: HTMLCell.topleft
        """
        return False


class HTMLCell(Cell):
    """ The Cell __init__ signature is:
    def __init__(self, value=None, column=None, type=None):
    where 'value' is the primary input, 'column' is a column name, and
    type is messytables.types.StringType() or better."""

    def __init__(self, value=None, column=None, type=None, source=None):
        assert value is None
        assert isinstance(source, lxml.etree._Element)
        self._lxml = source
        if type is None:
            from messytables.types import StringType
            type = StringType()
        self.type = type
        self.column = column
        self.column_autogenerated = False

    @property
    def topleft(self):
        """
        HTMLCells are those which are physically present in the HTML. They are
        always the top-left in their span.

        See also: FakeHTMLCell.topleft
        """
        return True

    @property
    def value(self):
        return text_from_element(self._lxml)

    @property
    def properties(self):
        return HTMLProperties(self._lxml)


def text_from_element(elem):
    r"""
    >>> x = lxml.html.fromstring('''<td><center><span style="display:none; speak:none" class="sortkey">01879-07-01-0000</span>
    ...                             <span style="white-space:nowrap;">1 July 1879</span></center></td>''')
    >>> text_from_element(x)
    '1 July 1879'
    >>> y = lxml.html.fromstring('''<td>a<br>b<br>c<br></td>''')
    >>> text_from_element(y)
    'a\nb\nc'
    """
    builder = []
    for x in elem.iter():
        #print x.tag, x.attrib, x.text, x.tail
        if is_invisible_text(x):
            cell_str = x.tail or ''  # handle None values.
        else:
            cell_str = (x.text or '') + (x.tail or '')
        cell_str = cell_str.replace('\n', ' ').strip()
        if x.tag == 'br' or x.tag == 'p':
            cell_str = '\n' + cell_str
        builder.append(cell_str)
    return ''.join(builder).strip()


def is_invisible_text(elem):
    flag = False
    if elem.tag == "span":
        if 'style' in elem.attrib:
            if 'display:none' in elem.attrib['style']:
                flag = True

    return flag


class HTMLProperties(CoreProperties):
    KEYS = ['_lxml', 'html', 'colspan', 'rowspan']

    def __init__(self, lxml_element):
        if not isinstance(lxml_element, lxml.etree._Element):
            raise TypeError("%r" % lxml_element)
        super(HTMLProperties, self).__init__()
        self.lxml_element = lxml_element

    def get_html(self):
        return lxml.html.tostring(self.lxml_element)

    def get__lxml(self):
        return self.lxml_element

    def get_colspan(self):
        try:
            return int(self.lxml_element.attrib.get('colspan', 1))
        except ValueError:
            return 1

    def get_rowspan(self):
        try:
            return int(self.lxml_element.attrib.get('rowspan', 1))
        except ValueError:
            return 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from . import horror_fobj
from messytables.any import any_tableset
from messytables.core import RowSet
from messytables.error import TableError
try:
    from nose.tools import assert_is_instance
except ImportError:
    from shim26 import assert_is_instance


class TestTableSet(unittest.TestCase):
    def setUp(self):
        fh = horror_fobj('simple.xls')
        self.table_set = any_tableset(fh, extension='xls')

    def test_get_item(self):
        assert_is_instance(self.table_set['simple.csv'], RowSet)

    def test_missing_sheet(self):
        self.assertRaises(TableError, lambda: self.table_set['non-existent'])

        # TODO: It would be good if we could manipulate a tableset to have
        # multiple row sets of the same name, then enable the following test.

        # self.assertRaises(Error, lambda: table_set['duplicated-name'])



from messytables.core import RowSet, TableSet, Cell

from messytables.types import StringType

try:
    from pdftables import get_tables
except ImportError as exc:
    if "No module named" not in exc.args[0]:
        raise
    get_tables = None


class PDFCell(Cell):

    def __init__(self, pdftables_cell):

        self._cell = pdftables_cell

        if pdftables_cell.topleft:
            w, h = pdftables_cell.size
            self._properties = dict(
                colspan=w,
                rowspan=h,
            )
            self.value = pdftables_cell.content

        else:
            self._properties = {}
            self.value = ""

        self.column = None
        self.column_autogenerated = False
        self.type = StringType()

    @property
    def topleft(self):
        return self._cell.topleft

    @property
    def properties(self):
        return self._properties


class PDFTableSet(TableSet):
    """
    A TableSet from a PDF document.
    """
    def __init__(self, fileobj=None, filename=None, **kw):
        if get_tables is None:
            raise ImportError("pdftables is not installed")
        if filename is not None:
            self.fh = open(filename, 'r')
        elif fileobj is not None:
            self.fh = fileobj
        else:
            raise TypeError('You must provide one of filename or fileobj')
        self.raw_tables = get_tables(self.fh)

    def make_tables(self):
        """
        Return a listing of tables (as PDFRowSets) in the table set.
        """
        def table_name(table):
            return "Table {0} of {1} on page {2} of {3}".format(
                table.table_number_on_page,
                table.total_tables_on_page,
                table.page_number,
                table.total_pages)
        return [PDFRowSet(table_name(table), table)
                for table in self.raw_tables]


class PDFRowSet(RowSet):
    """
    A RowSet representing a PDF table.
    """
    def __init__(self, name, table):
        if get_tables is None:
            raise ImportError("pdftables is not installed")
        super(PDFRowSet, self).__init__()
        self.name = name
        self.table = table
        self.meta = dict(
            page_number=table.page_number + 1,
        )

    def raw(self, sample=False):
        """
        Yield one row of cells at a time
        """
        if hasattr(self.table, "cell_data"):
            # New style of cell data.
            for row in self.table.cell_data:
                yield [PDFCell(pdf_cell) for pdf_cell in row]
        else:
            for row in self.table:
                yield [Cell(pdf_cell) for pdf_cell in row]



class MessytablesError(Exception):
    """A generic error to inherit from"""


class ReadError(MessytablesError):
    '''Error reading the file/stream in terms of the expected format.'''
    pass


class TableError(MessytablesError, LookupError):
    """Couldn't identify correct table."""
    pass

class NoSuchPropertyError(MessytablesError, KeyError):
    """The requested property doesn't exist"""
    pass



import sys
PY2 = sys.version_info[0] == 2
if PY2:
    import urllib2
    from itertools import izip_longest
    unicode_string = unicode
    native_string = str
    byte_string = str
    string_types = (str, unicode)
    urlopen = urllib2.urlopen
else:  # i.e. PY3
    import urllib.request
    from itertools import zip_longest as izip_longest
    unicode_string = str
    native_string = str
    byte_string = bytes
    
    string_types = (str,)
    urlopen = urllib.request.urlopen
